/************************************************************************/
/*									*/
/*									*/
/*	This file contains pid descriptor management routines		*/
/*									*/
/*									*/
/************************************************************************/

/*************** Includes ***************************/

#include "xwio.h"

/************** Pidlist definitions ***********************/

/*
 *	pidlist is an array of data reference descriptors 'pidref'.
 *	The first entry in this table (plist[0]) is a link to further tables
 *	allowing more than PIDLIST_COUNT process active. (maximum number
 *	of pids is 65280).
 *	The second word of this 'link entry' in each table contains a lru
 *	offset. This is the offset of the last entry used. A new pid is placed
 *	in the table after this entry, if possible.
 *	A pid consists of an offset in a table, a table count and a usage
 *	counter.
 *	In unused entries in the table offset and table count are set
 *	to zero.
 *	The usage counter is incremented with every new pid in this entry.
 *	The 'data'entry in a pidref struct contains a pointer to any type of
 *	object connected to a unique pid.
 *	In unused entries, the data pointer is set to NULL.
 *
 *	plist_add and plist_del run in HiPri to ensure collision free
 *	processing. (Alternatively,  there may be a semaphore...)
 *
 *	The base of the table list must be known globally, preferably by
 *	writing plist into every GD.
 *
 *	The standard use of pidlists is the management of process descriptors.
 *	Multiple pid lists may be used to control other data to which
 *	unique handles are needed, such as objects...
 *	Alternatively, several data types can 'share' a single pid list,
 *	if the type of the data struct can be recognized by a 'magic' number.
 *
 */


#define PIDLIST_SIZE	       256
#define PIDLIST_SIZE_MASK      0x000000FF
#define PIDREF_COUNT_MASK      0x0FFF0000   /* use 28bit only (store in INT) */
#define PIDREF_REF_MASK        0x0000FFFF
#define PIDREF_TABLE_MASK      0x0000FF00
#define PIDREF_OFFSET_MASK     0x000000FF
#define PIDLIST_USAGE_INC      0x00010000

struct pidref *plist_init ()
/*
 *	initialize a node-global pid list
 */

	{
	struct pidref *plist;
	int i;

	/*ERROR ("initializing plist.\n");*/
	plist=(struct pidref *)malloc (PIDLIST_SIZE*sizeof (struct pidref));
	for (i=0; i<PIDLIST_SIZE; i++) {
		plist[i].pid=0;
		plist[i].data=NULL;
		}
	/*ERROR ("returning from plist_init: %X.\n",plist);*/
	return plist;
	}

int plist_add(struct pidref *plist, void *data)
/*
 *	store descriptor in pid list, return pid
 */
	{
	int i, pid, lru;
	int table=0;
	int done=0;
	struct pidref *np;

	do	{
		lru=(int)plist[0].data;
		for (i=1; i<PIDLIST_SIZE; i++)
			if (!(plist[(i+lru)&PIDLIST_SIZE_MASK].pid & PIDREF_REF_MASK) && ((i+lru)&PIDLIST_SIZE_MASK))
				break;
		if (i==PIDLIST_SIZE) {	       /* table is full */
			if (!plist[0].pid) {	/* make new table */
				np=plist_init ();
				plist[0].pid=(int)np;
				}
			plist=(struct pidref *)plist[0].pid;
			table++;
			}
		else	{			/* found empty entry */
			i=(i+lru) & PIDLIST_SIZE_MASK;
			(int)plist[i].data=i;/* update lru */
			plist[i].pid=(plist[i].pid+PIDLIST_USAGE_INC)&PIDREF_COUNT_MASK;
			if (!plist[i].pid)
				/* usage counter overflow */
				;
			plist[i].pid|=((table<<8) & PIDREF_TABLE_MASK)|(i & PIDREF_OFFSET_MASK);
			plist[i].data=data;
			done=1;
			}
		}while (!done);

	return plist[i].pid;
	}

int plist_next (struct pidref *plist, int pid)
/*
 *	return next valid pid in plist or NULL, if pid is NULL, return first.
 *	This allows stepping through the list for cleanup...
 */
	{
	int i, table;
	if (pid)
		{
		table=pid & PIDREF_TABLE_MASK;
		while (table--)
			plist=(struct pidref *)plist[0].pid;
		i=pid & PIDREF_OFFSET_MASK;
		}
	else	{
		i=0;
		table=0;
		}
	do	{
		i++;
		if (i==PIDLIST_SIZE)
			{
			if (plist[0].pid)
				{
				plist=(struct pidref *)plist[0].pid;
				table++;
				i=1;
				}
			else	{
				return 0;
				}
			}
		} while (!plist[i].data);
	return plist[i].pid;
	}

void *plist_get (struct pidref *plist, int pid)
/*
 *	find data for a given pid
 */
	{
	int i;
	int table;

	table=pid & PIDREF_TABLE_MASK;
	while (table--)
		plist=(struct pidref *)plist[0].pid;
	i=pid & PIDREF_OFFSET_MASK;
	if (plist[i].pid!=pid){
		ERROR("PLIST_GET: PID out of date!");
		}
	return plist[i].data;
	}

void *plist_del(struct pidref *plist, int pid)
/*
 *	find pid, remove entry, return data pointer
 */

	{
	int i, table;
	void *p;

	table=pid & PIDREF_TABLE_MASK;
	while (table--)
		plist=(struct pidref *)plist[0].pid;
	i=pid & PIDREF_OFFSET_MASK;
	if (plist[i].pid!=pid){
		ERROR("PLIST_DEL: PID out of date!");
		}
	plist[i].pid &= PIDREF_COUNT_MASK;     /* delete reference data, keep counter */
	p=plist[i].data;
	plist[i].data=NULL;

	return p;
	}

